package imageMsg;

import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import javax.swing.*;
import javax.swing.filechooser.FileFilter;

import javax.imageio.*; 
import java.awt.event.*;

public class BitFlipper {

	//panel for displaying errors:
	final static JTextPane errText = new JTextPane();

	//add files must be treated as png for encoding purposes
	public static String extension(String fileName){
		return "png";
		//int extStart = fileName.lastIndexOf(".");
		//return fileName.substring(extStart+1,fileName.length());
	}
					
	public static boolean encode(String inImg, String outImg, String msg){
		//spaces are ignored to improve encoding efficiency
		msg = msg.toLowerCase().replaceAll(" ","");
		if(msg.length() == 0){
			errText.setText("Empty message not allowed");
			System.err.println("Must encode non-empty message");
			return false;
		}
		try{
			System.out.println("Starting Encoding...");
			
			File infile = new File(inImg); 
			BufferedImage im = 
			ImageIO.read(infile);
			
			int colIndex = im.getWidth() / 26;
			int rowIndices = im.getHeight() / msg.length();			
			
			for(int index = 0; index < msg.length(); ++index){
				int row = rowIndices * index + (int) (Math.random() * (rowIndices - 1));
				int col = ((msg.charAt(index) - 'a') * colIndex) + ((int) (Math.random() * (colIndex-1)));
				if(col < 0 || col > im.getWidth()){
					//only letters are encoded
					errText.setText("Invalid (non-alpha) characters in message");
					System.err.println("Invalid (non-alpha) characters in message");
					return false;
				}
				if(im.getRGB(col,row)%2 == 0){
					im.setRGB(col,row,im.getRGB(col,row)+1);
				}else{
					im.setRGB(col,row,im.getRGB(col,row)-1);
				}
			}
						
			System.out.println("Encoding complete.");

			File outfile = new File(outImg);
			ImageIO.write(im, extension(outImg), outfile);
			return true;
		}
		catch(IOException ioe){
			errText.setText("IO Exception, possibly missing file");
			System.err.println("Io exception");
			return false;
		}

	}
	
	//method to extract a hidden message from stego object also given cover
	public static String decode(String orgImg, String codeImg){
		String msg = "";
		try{
			System.out.println("Starting decoding...");
			
			File plainFile = new File(orgImg); 
			BufferedImage plain = 
			ImageIO.read(plainFile);
			
			File codedFile = new File(codeImg); 
			BufferedImage coded = 
			ImageIO.read(codedFile);
			
			if(!(coded.getWidth() == plain.getWidth() && coded.getHeight() == plain.getHeight())){
				errText.setText("Incompatible images for decode");
				System.err.println("Incompatible images");
				return msg;
			}
						
			int colIndex = coded.getWidth() / 26;
			
			for(int row = 0; row < coded.getHeight(); ++row){
				for(int col = 0; col < coded.getWidth(); ++col){
					if(plain.getRGB(col,row) != coded.getRGB(col,row)){
						//should be only one per row
						char thisChar = (char) ('a' + (int) (col/colIndex));
						msg += thisChar;
					}
				}
			}
			
			System.out.println("Decoding complete.");
						
			return msg;
		}
		catch(IOException ioe){
			errText.setText("IO Exception, possibly missing file");
			System.err.println("Io exception");
		}
		return msg;
	}
	
	//method to add noise to inFile and output noisy file as outFile
	public static void randomNoise(String inFile, String outFile){
		try{
			System.out.println("Starting Generation...");
			
			File infile = new File(inFile); 
			BufferedImage im = 
			ImageIO.read(infile);	
			
			for(int row = 0; row < im.getHeight(); ++row){
				for(int col = 0; col < im.getWidth(); ++col){
					if(Math.random() > 0.5){
						if(im.getRGB(col,row)%2 == 0){
							im.setRGB(col,row,im.getRGB(col,row)+1);
						}else{
							im.setRGB(col,row,im.getRGB(col,row)-1);
						}
					}
				}
			}
						
			System.out.println("Generation complete.");

			File outfile = new File(outFile);
			ImageIO.write(im, extension(outFile), outfile);
		}
		catch(IOException ioe){
			errText.setText("IO Exception, possibly missing file");
			System.err.println("Io exception");
		}
	}
	
	//method to generate a fully random image and save it as fileName
	public static void randomImage(String fileName){
		try{
			System.out.println("Starting Generation...");
			
			File infile = new File("blank.png"); 
			BufferedImage im = 
			ImageIO.read(infile);	
			
			for(int row = 0; row < im.getHeight(); ++row){
				for(int col = 0; col < im.getWidth(); ++col){
					im.setRGB(col,row,(int)(Math.random()*Math.pow(2,32)));
				}
			}
						
			System.out.println("Generation complete.");

			File outfile = new File(fileName);
			ImageIO.write(im, extension(fileName), outfile);
		}
		catch(IOException ioe){
			errText.setText("IO Exception, possibly missing file");
			System.err.println("Io exception");
		}
	}
	
	//method to generate a semi-Random image and save it as file name
	public static void semiRandomImage(String fileName){
		try{
			System.out.println("Starting Generation...");
			
			File infile = new File("blank.png"); 
			BufferedImage im = 
			ImageIO.read(infile);		
			
			for(int row = 0; row < im.getHeight(); ++row){
				for(int col = 0; col < im.getWidth(); ++col){
					if(row == 0 || col == 0){
						im.setRGB(col,row,(int)(Math.random()*Math.pow(2,32)));
					}else{
						int base = im.getRGB(col-1,row);
						if(Math.random() < 0.5){
							base = im.getRGB(col,row-1);
							
						}
						im.setRGB(col,row,base - 5 + (int)(Math.random() * 10));
					}
				}
			}
						
			System.out.println("Generation complete.");

			File outfile = new File(fileName);
			ImageIO.write(im, extension(fileName), outfile);
		}
		catch(IOException ioe){
			errText.setText("IO Exception, possibly missing file");
			System.err.println("Io exception");
		}
	}
	
	//method to count pixels
	public static int[] evenOddCounter(String fileName){
		int[] evensOdds = new int[3];
		try{
			System.out.println("Starting counting...");
			
			File file = new File(fileName); 
			BufferedImage img = ImageIO.read(file);
			
			int prevPixel = 0;
			int thisPixel = 0;
			for(int row = 0; row < img.getHeight(); ++row){
				for(int col = 0; col < img.getWidth(); ++col){
					thisPixel = img.getRGB(col,row);
					++(evensOdds[Math.abs(thisPixel%2)]);
					if(thisPixel % 2 != prevPixel %2){
						++(evensOdds[2]); //count how often it changes.
					}
					prevPixel = thisPixel;
				}
			}
			
		}catch(IOException ioe){
			errText.setText("IO Exception, possibly missing file");
			System.err.println("IO Exception");
		}
		return evensOdds;
	}
	
	//method to calculate and return statistics comparing cover and stego
	public static String stats(String input, String output){
		int[] originalNums = evenOddCounter(input);
		int[] codedNums = evenOddCounter(output);
		
		String results = "Even pixels:  " + originalNums[0] + " original and " + codedNums[0] + " encoded";
		results += "\n" + "Odd pixels:  " + originalNums[1] + " original and " + codedNums[1] + " encoded";
		results += "\n" + "Switching pixels:  " + originalNums[2] + " original and " + codedNums[2] + " encoded";
		return results;

	}

	public static void main(String[] args){
		
		//interface for users:
		UISetup();
		
		//some sample runs of the program apart from the UI:
		String message = "We will attack at dawn  Pass this message on to your generals and to no one else I add even more to confuse the matter further";

		String input = "penny.jpg";
		String inNoise = "penny-noise.jpg";
		
		randomNoise(input,inNoise);
		for(int i = 1; i < 10; ++i){ //many iterations of noise
			randomNoise(inNoise,inNoise);
		}
		
		String output = "penny-after.jpg";
		String outNoise = "penny-noise-after.jpg";
		
		encode(input, output, message);
		System.out.println("decoded: " + decode(input,output));
		
		stats(input,output);
		
		encode(inNoise, outNoise, message);
		System.out.println("decoded: " + decode(inNoise,outNoise));
		
		stats(inNoise,outNoise);
		
			
		String randImg = "testa.png";
		String randOut = "testb.png";
		randomImage(randImg);
		encode(randImg, randOut, message);
		System.out.println("decoded: " + decode(randImg,randOut));
		
		stats(randImg, randOut);
	
		
		/*String semiRandImg = "s-testa.png";
		String semiRandOut = "s-testb.png";
		semiRandomImage(semiRandImg);
		encode(semiRandImg, semiRandOut, message);
		System.out.println("decoded: " + decode(semiRandImg,semiRandOut));
		
		stats(semiRandImg, semiRandOut);*/
	}
	
	//method to generate UI
	//Developers may create their own UI using the algorithms/methods defined above
	public static void UISetup(){
		
		final String fullPath = "";
		
		JFrame frame = new JFrame("C-Hide");
		final JPanel panel = new JPanel();
		
		panel.setLayout(new BoxLayout(panel, BoxLayout.Y_AXIS));
		
		JButton fileButton = new JButton("Choose File");
		final JLabel openedFile = new JLabel("[none]");
		fileButton.addActionListener(new ActionListener(){

			public void actionPerformed(ActionEvent arg0) {
			    JFileChooser chooser = new JFileChooser();

				//only allow (some) image file types
			    FileFilter filter = new FileFilter(){

					public boolean accept(File arg0) {
						String fileName = arg0.getName().toLowerCase();
						if(fileName.endsWith(".jpg") || fileName.endsWith(".jpeg") || 
								fileName.endsWith(".bmp") || fileName.endsWith(".gif") ||
								fileName.endsWith(".png") || arg0.isDirectory()){
							return true;
						}
						return false;
					}

					public String getDescription() {
						return "*.jpg, *.gif, *.bmp, *.png";
					}
			    	
			    };
			    chooser.setFileFilter(filter);
			    int returnVal = chooser.showOpenDialog(panel);
			    if(returnVal == JFileChooser.APPROVE_OPTION) {
			       openedFile.setText(chooser.getSelectedFile().getAbsolutePath());
			    }
				
			}
			
		});
		
		//message entry
		JLabel lab1 = new JLabel("Enter Text To Hide: ");
		final JTextField tb1 = new JTextField();
		tb1.setSize(1,100);
		
		//technique selection
		final ButtonGroup bg1 = new ButtonGroup();
		final JRadioButton normal = new JRadioButton("normal");
		final JRadioButton noise = new JRadioButton("noise");
		final JRadioButton random = new JRadioButton("random");
		final JRadioButton semiRandom = new JRadioButton("semi-random");
		
		bg1.add(normal);
		bg1.add(noise);
		bg1.add(random);
		bg1.add(semiRandom);
		
		JButton goButton = new JButton("Go!");
		
		final JLabel newFile = new JLabel("[none]");
		
		JLabel stats = new JLabel("Statistics:");
		final JTextPane tb2 = new JTextPane();
		tb2.setEditable(false);
		JLabel errors = new JLabel("Errors:");
		final JTextPane tb3 = errText;
		tb3.setEditable(false);
		
		//operation of algorithms:
		goButton.addActionListener(new ActionListener(){

			public void actionPerformed(ActionEvent arg0) {
				tb3.setText("");
				try{
					boolean encoded = false;
					String input = openedFile.getText();
					String output = ""; 
					//output could be calculated here, but seems better to only calculate
					//if a valid radio button is selected
					String message = tb1.getText();
					if(input == ""){
						tb3.setText("Must specify a file");
					}
					if(normal.isSelected()){
						int extStart = input.lastIndexOf(".");
						output = input.substring(0,extStart) + "-o" + input.substring(extStart,input.length());
						
						encoded = encode(input, output, message);
					}
					else if(noise.isSelected()){
						int extStart = input.lastIndexOf(".");
						output = input.substring(0,extStart) + "-o" + input.substring(extStart,input.length());
						
						randomNoise(input,output);
						for(int i = 1; i < 10; ++i){ //many iterations of noise
							randomNoise(output,output);
						}
						encoded = encode(input, output, message);
					}
					else if(random.isSelected()){
						
						int extStart = input.lastIndexOf(".");
						output = input.substring(0,extStart) + "-o" + input.substring(extStart,input.length());

						randomImage(input);
						encoded = encode(input, output, message);
					}
					else if(semiRandom.isSelected()){
						
						int extStart = input.lastIndexOf(".");
						output = input.substring(0,extStart) + "-o" + input.substring(extStart,input.length());
						
						semiRandomImage(input);
						encoded = encode(input, output, message);
					}
					else{
						tb3.setText("Must select encoding type");//error
					}
					if(encoded){
						//if encoding failed, we don't update here
						newFile.setText(output);
						tb2.setText(stats(input,output));
					}	
				}
				catch(StringIndexOutOfBoundsException e){
					tb3.setText("Invalid message to encode");
				}
			}
			
		});
		
		JButton decodeButton = new JButton("Decode");
		final JTextPane tb4 = new JTextPane();
		
		//display results of decoding:
		decodeButton.addActionListener(new ActionListener(){

			public void actionPerformed(ActionEvent arg0) {
				String input = openedFile.getText();
				int extStart = input.lastIndexOf(".");
				String output = input.substring(0,extStart) + "-o" + input.substring(extStart,input.length());
				tb4.setText(decode(input,output));
			}
			
		});

		
		//build the panel:
		panel.add(fileButton);
		panel.add(openedFile);
		panel.add(lab1);
		panel.add(tb1);
		panel.add(normal);
		panel.add(noise);
		panel.add(random);
		panel.add(semiRandom);
		panel.add(goButton);
		panel.add(newFile);
		panel.add(stats);
		panel.add(tb2);
		panel.add(errors);
		panel.add(tb3);
		panel.add(decodeButton);
		panel.add(tb4);
		
		
		frame.add(panel);
		
		//show it:
		frame.pack();
		frame.setVisible(true);
	}
	
}
